// Boss.cpp
// Brian Seney, Alfredo Porras
// IMGD3000

#include "Boss.h"

#include "Engine/ResourceManager.h"
#include "Engine/GraphicsManager.h"
#include "Engine/WorldManager.h"
#include "Engine/EventManager.h"
#include "Engine/EventStep.h"
#include "Engine/EventCollision.h"
#include "EventTargetMovement.h"

#include "EnemyArrow.h"
#include "Opponent.h"
#include "OpponentZombie.h"
#include "Goal.h"
#include "Picture.h"
#include "EventScore.h"

#define NUM_STATES 2

Boss::Boss(){
	object_type = OPPONENT_TYPE;

	health = 5;

	sprite_slowdown = 20;
	sprite_countdown = sprite_slowdown;
	sprite_index = 0;

	state = 0;
	state_slowdown = 1000;
	state_countdown = state_slowdown;

	move_slowdown = 20;
	move_countdown = move_slowdown;

	target = Position(-1, -1);

	WorldManager &world_manager = WorldManager::getInstance();
	world_manager.addObj(this);

	EventManager &event_manager = EventManager::getInstance();
	event_manager.registerInterest(this, STEP_EVENT);
	event_manager.registerInterest(this, TARGETMOVEMENT_EVENT);
	event_manager.registerInterest(this, COLLISION_EVENT);

	ResourceManager &resource_manager = ResourceManager::getInstance();
	sprite_ptr = resource_manager.loadSprite("Sprites/Boss.txt");
	obj_box = Box(pos, sprite_ptr->getWidth(), sprite_ptr->getHeight());
}

Boss::~Boss(){
	EventManager &event_manager = EventManager::getInstance();
	event_manager.removeInterest(this, STEP_EVENT);
	event_manager.removeInterest(this, TARGETMOVEMENT_EVENT);
	event_manager.removeInterest(this, COLLISION_EVENT);
}

int Boss::eventHandler(Event *event_ptr){
	if(event_ptr->getType().compare( STEP_EVENT ) == 0){
		// determine whether to switch states
		state_countdown--;
		if(state_countdown <= 0){
			state_countdown = state_slowdown;
			state++;
			if(state >= NUM_STATES){
				state = 0;
			}
		}
		// state 0: follow player
		if(state == 0){
			move_countdown--;
			if(move_countdown <= 0){
				move_countdown = move_slowdown;
				// move closer to the target, if the target exists
				// if at the target, set the target at -1, -1
				if(target.get_x() != -1){
					if( pos == target){
						target = Position(-1, -1);
						return 1;
					}
					if( abs(pos.get_x() - target.get_x()) >
					    abs(pos.get_y() - target.get_y())){
						if(pos.get_x() > target.get_x()){
							move(-1, 0);
						} else {
							move(1, 0);
						}
					} else {
						if(pos.get_y() > target.get_y()){
							move(0, -1);
						} else {
							move(0, 1);
						}
					}
				}
			// if halfway to the next move, attack		
			} else if(move_countdown * 2 == move_slowdown){
				attack();
			}
		// state 1: stay stationary, summon enemies
		} else if(state == 1){
			move_countdown--;
			if(move_countdown + 200 <= move_slowdown){
				move_countdown = move_slowdown;
				spawnZombies();
			}
		}
		return 1;
	} else if(event_ptr->getType().compare( TARGETMOVEMENT_EVENT ) == 0){
		EventTargetMovement *event_target_ptr = static_cast <EventTargetMovement *> (event_ptr);
		Position temp_target = event_target_ptr->getTargetPos();
		target = Position(temp_target.get_x() - 4, temp_target.get_y() - 3);
		return 1;
	} else if (event_ptr->getType() == COLLISION_EVENT) {
		EventCollision *collision_ptr = static_cast<EventCollision*>(event_ptr);
		GameObject* obj_ptr = collision_ptr->getObj2();
		// If colliding with the ball, take damage.
		if(obj_ptr->getType().compare( BALL_TYPE ) == 0){
			// Destroy the ball; send the score event
			WorldManager &world_manager = WorldManager::getInstance();
			world_manager.markForDelete(obj_ptr);

			EventManager &event_manager = EventManager::getInstance();
			event_manager.sendEvent( new EventScore(1) );

			// Display the goal sprite
			ResourceManager &resource_manager = ResourceManager::getInstance();
			Sprite *sprite_ptr = resource_manager.loadSprite("Sprites/GoalString.txt");
			Picture *temp = new Picture(sprite_ptr);
			temp->setPos(11, 8);

			health -= 1;
			if(health <= 0){
				WorldManager &world_manager = WorldManager::getInstance();
				world_manager.markForDelete(this);
				
				Position goal_pos = Position(56, 8);
				// Destroy the wall in front of the goal
				GameObjectList obj_list = world_manager.getAllObjects();
				list<GameObject*>::iterator iter = obj_list.begin();
				while(iter != obj_list.end() ){
					Position temp_pos = (*iter)->getPos();
					if(goal_pos.get_x() <= temp_pos.get_x() &&
					   temp_pos.get_x() < goal_pos.get_x() + 3 &&
					   goal_pos.get_y() <= temp_pos.get_y() &&
					   temp_pos.get_y() < goal_pos.get_y() + 6){
						world_manager.markForDelete(*iter);
					}
					iter++;
				}
				// Create a goal
				Goal* goal_ptr = new Goal();
				goal_ptr->setPos(goal_pos);
					
			}
		}
		return 1;
	}
	return 0;
}

void Boss::draw(){
	GraphicsManager &graph_manager = GraphicsManager::getInstance();
	// Find the frame to draw
	sprite_countdown--;
	// If ready to draw a new frame, increase the index
	if(sprite_countdown < 0){
		sprite_countdown = sprite_slowdown;
		sprite_index++;
		if(sprite_index >= sprite_ptr->getNumFrames() ){
			sprite_index = 0;
		}
	}
	graph_manager.drawFrame(pos, sprite_ptr->getFrame(sprite_index));

	// Drawing health bar
	string health_bar = "Boss Health:";
	int temp_health = health;
	while(temp_health > 0){
		temp_health = temp_health - 1;
		health_bar.append("XX");
	}
	char temp[50];
	strcpy(temp, health_bar.c_str());
	graph_manager.drawStr(Position(17, 1), temp);

	// Draw the instructions bar
	string instruction_bar = "Hit it with the ball!";
	char temp2[50];
	strcpy(temp2, instruction_bar.c_str());
	graph_manager.drawStr(Position (18, 22), temp2);
}

void Boss::attack(){
	new EnemyArrow(pos.get_x() - 1, pos.get_y() - 1, -0.8, -0.8);
	new EnemyArrow(pos.get_x() - 1, pos.get_y() + 2, -0.8, 0);
	new EnemyArrow(pos.get_x() - 1, pos.get_y() + 5, -0.8, 0.8);

	new EnemyArrow(pos.get_x() + 3, pos.get_y() - 1,  0, -0.8);
	new EnemyArrow(pos.get_x() + 3, pos.get_y() + 5,  0, 0.8);

	new EnemyArrow(pos.get_x() + 7, pos.get_y() - 1,  0.8, -0.8);
	new EnemyArrow(pos.get_x() + 7, pos.get_y() + 2,  0.8, 0);
	new EnemyArrow(pos.get_x() + 7, pos.get_y() + 5,  0.8, 0.8);
}

void Boss::spawnZombies(){
	OpponentZombie* zombie_ptr = new OpponentZombie();
	zombie_ptr->setPos(pos.get_x() - 1, pos.get_y() - 1);
	zombie_ptr = new OpponentZombie();
	zombie_ptr->setPos(pos.get_x() - 1, pos.get_y() + 5);
	zombie_ptr = new OpponentZombie();
	zombie_ptr->setPos(pos.get_x() + 7, pos.get_y() - 1);
	zombie_ptr = new OpponentZombie();
	zombie_ptr->setPos(pos.get_x() + 7, pos.get_y() + 5);
}

void Boss::move(int x, int y){
	WorldManager &world_manager = WorldManager::getInstance();
	// if stuck on something, find a new target
	if(world_manager.moveObj(this, Position(pos.get_x() + x, pos.get_y() + y)))
		target = Position(-1, -1);
}
