// GameManager.cpp
// Brian Seney
// IMGD3000

#include "GameManager.h"
#include "LogManager.h"
#include "WorldManager.h"
#include "InputManager.h"
#include "GraphicsManager.h"
#include "EventManager.h"
#include "ResourceManager.h"
#include <unistd.h>

// Start the game
int GameManager::startUp(int debug, bool append){
	// If the manager is already running, don't start again.
	if(is_running){
		printf("Warning: GameManager::startUp called while already running!\n");
		return 1;
	}

	// Initialize the managers
	LogManager &log_manager = LogManager::getInstance();
	log_manager.startUp(debug, append);

	GraphicsManager &graphics_manager = GraphicsManager::getInstance();
	graphics_manager.startUp();

	InputManager &input_manager = InputManager::getInstance();
	input_manager.startUp();

	WorldManager &world_manager = WorldManager::getInstance();
	world_manager.startUp();

	EventManager &event_manager = EventManager::getInstance();
	event_manager.startUp();

	ResourceManager &resource_manager = ResourceManager::getInstance();
	resource_manager.startUp();

	// Initialize the frame rate (default of 30)
	frame_rate = 30;

	is_started = true;
	// Print an initialization message
	log_manager.writeLog(1, "Game Manager: Started!");
	return 0;
}

// Shutdown the game
int GameManager::shutDown(){
	// Shut down the managers
	WorldManager &world_manager = WorldManager::getInstance();
	world_manager.shutDown();

	InputManager &input_manager = InputManager::getInstance();
	input_manager.shutDown();

	GraphicsManager &graphics_manager = GraphicsManager::getInstance();
	graphics_manager.shutDown();

	LogManager &log_manager = LogManager::getInstance();
	log_manager.shutDown();

	EventManager &event_manager = EventManager::getInstance();
	event_manager.shutDown();

	ResourceManager &resource_manager = ResourceManager::getInstance();
	resource_manager.shutDown();

	is_started = false;

	return 0;
}

// Start the game loop
void GameManager::run(){
	WorldManager &world_manager = WorldManager::getInstance();
	GraphicsManager &graphics_manager = GraphicsManager::getInstance();
	InputManager &input_manager = InputManager::getInstance();

	// Initialize the clock
	Clock* clock_ptr = new Clock();
	is_running = true;
	// How many microseconds each step should take
	long int frame_ticks = FRAME_TICKS;

	while(is_running){
		// If the loop ran too quickly, wait to
		// synchronize the frame rate
		long int temp_t = clock_ptr->delta_t();
		if(temp_t < frame_ticks){
			usleep(frame_ticks - temp_t);
		}

		// Obtain any input
		input_manager.getInput();

		// Update and draw the world
		world_manager.update(); // also draws all objects
		graphics_manager.swapBuffers();

		// And run forever, unless something changes is_running
	}
	delete clock_ptr;
}

// Stop the game loop
void GameManager::stop(){
	is_running = false;
}
