// GraphicsManager.cpp
// Brian Seney
// IMGD3000

#include "GraphicsManager.h"
#include "GameManager.h"
#include "LogManager.h"
#include "Frame.h"

#include <vector>

/* Colors defined as:
COLOR_BLACK   0
COLOR_RED     1
COLOR_GREEN   2
COLOR_YELLOW  3
COLOR_BLUE    4
COLOR_MAGENTA 5
COLOR_CYAN    6
COLOR_WHITE   7 */

int GraphicsManager::startUp(){
//	is_started = true;
	LogManager &log_manager = LogManager::getInstance();
	log_manager.writeLog(1, "Graphics Manager: Started!");

	initscr();
	getmaxyx(stdscr, window_y, window_x);
	win1 = newwin(0,0,0,0);
	win2 = newwin(0,0,0,0);
	current_window = win1;
	leaveok(stdscr, TRUE);

	if (has_colors() == TRUE) {
		log_manager.writeLog(1, "Terminal supports colors.");
		start_color();
	}
	else {
		log_manager.writeLog(1, "Terminal does not support colors.");
	}
	wattron(stdscr, A_BOLD);

	is_started = true;

	// Initialize the color pairs for the window
	// Sets 64 color pairs from 0-63. For example.
	// pair 0 = background 0, foreground 0
	// pair 1 = background 0, foreground 1
	// pair 2 = background 0, foreground 2
		// ..
	// pair 7 = background 0, foreground 7
	// pair 8 = background 1, foreground 0
		// etc up to pair 63
	for(int i = 0; i < 8; i++){
		for(int j = 0; j < 8; j++){
			init_pair(j + (i * 8), j, i);
		}
	}

	// initalize the camera
	camera = Position(0, 0);

	return 0;
}

int GraphicsManager::shutDown(){
	is_started = false;
	// Shut down curses
	delwin(win1);
	delwin(win2);
	endwin();

	return 0;

}

Position GraphicsManager::worldToScreen(Position pos){
	//Position temp;
	pos.set_x( pos.get_x() + camera.get_x() );
	pos.set_y( pos.get_y() + camera.get_y() );
	return pos;
}

// Switch the current window, refresh the display, clear the other buffer
void GraphicsManager::swapBuffers(){
	LogManager &log_manager = LogManager::getInstance();
	if(wrefresh(current_window) == ERR) {
		log_manager.writeLog(1,"GraphicsManager::swapBuffers(): failed to refresh current window.");
	}
	if(current_window == win1) {
		if (werase(win2) == ERR) {
			log_manager.writeLog(1,"GraphicsManager::swapBuffers(): failed to erase win2 before switching to it.");
		}
		current_window = win2;
	}
	else {
		if (werase(win1) == ERR) {
			log_manager.writeLog(1,"GraphicsManager::swapBuffers(): failed to erase win1 before switching to it.");
		}
		current_window = win1;
	}
}

int GraphicsManager::drawCh(Position pos, char ch){
	if( mvwaddch(current_window, pos.get_y(), pos.get_x(), ch) )
		return 0;
	else
		return 1;
}

int GraphicsManager::drawCh(Position pos, char ch, int color){
	wattron(current_window, COLOR_PAIR(color));
	// translate the position based on the camera
	Position temp = worldToScreen(pos);
	if( mvwaddch(current_window, temp.get_y(), temp.get_x(), ch) ){
		wattroff(current_window, COLOR_PAIR(color));
		return 0;
	} else {
		wattroff(current_window, COLOR_PAIR(color));
		return 1;
	}
}

int GraphicsManager::drawStr(Position pos, char* ch){
	// translate the position based on the camera
	Position temp = worldToScreen(pos);
	if( mvwaddstr(current_window, temp.get_y(), temp.get_x(), ch) )
		return 0;
	else
		return 1;
}

int GraphicsManager::drawFrame(Position pos, Frame *frame_ptr){
	// if the pointer is NULL, return immediately
	if(!frame_ptr)
		return 1;

	int x = pos.get_x();
	int y = pos.get_y();

	int width = frame_ptr->getWidth() + x; // the boundaries to print to
	int height = frame_ptr->getHeight() + y;

	int i = 0; // index for both the character to draw and the color to draw it

	string print_string = frame_ptr->getString();
	vector<int> colors = frame_ptr->getColors();

	// start at the given position, and draw each character in the row
	for(y = pos.get_y(); y < height; y++){
		for(x = pos.get_x(); x < width; x++){
			drawCh(Position(x, y), print_string.at(i), colors[i]);
			i++;
		}
	}
	return 0;
}
