// WorldManager.cpp
// Brian Seney
// IMGD3000

#include "GameManager.h"
#include "GraphicsManager.h"
#include "WorldManager.h"
#include "EventManager.h"
#include "EventCollision.h"
#include "EventOut.h"

int WorldManager::startUp(){
	LogManager &log_manager = LogManager::getInstance();
	log_manager.writeLog(1, "World Manager: Started!");
	// By default the world view is equal to the screen view
	GraphicsManager &graph_manager = GraphicsManager::getInstance();
	world_boundary.setWidth(graph_manager.getMaxX() );
	world_boundary.setHeight(graph_manager.getMaxY() );

	for (int i = 0; i < MAP_WIDTH; i++) {
		for (int j = 0; j < MAP_HEIGHT; j++) {
			setPassable(i,j,true);
		}
	}

	is_started = true;
	clear = false;
	return 0;
}

int WorldManager::shutDown(){
	// Remove all objects in the lists
	game_objs.clear();
	rm_objs.clear();
	is_started = false;
	return 0;
}

int WorldManager::markForDelete(GameObject* obj_ptr){
	// Make sure the object is not already there, or update will delete
	// the object twice (which has undefined behavior)
	if(rm_objs.find(obj_ptr) == NULL){
		rm_objs.add(obj_ptr);
		return 0;
	} else {
		return 1;
	}
}

int WorldManager::addObj(GameObject* obj_ptr){
	// Make sure the object is not already there; this would make the
	// object accept updates twice
	if(game_objs.find(obj_ptr) == NULL){
		game_objs.add(obj_ptr);

		if(obj_ptr->isSolid()){
			setPassable(obj_ptr->getPos().get_x(),obj_ptr->getPos().get_y(), false);
		}

		LogManager &log_manager = LogManager::getInstance();
		log_manager.writeLog(2, "Object added to World.");

		return 0;
	} else {
		return 1;
	}
}

int WorldManager::removeObj(GameObject* obj_ptr){
	game_objs.remove(obj_ptr);

	LogManager &log_manager = LogManager::getInstance();
	log_manager.writeLog(2, "Object removed from World.");

	return 0;
}

int WorldManager::setPermanent(GameObject *obj_ptr){
	if(permanent_objs.find(obj_ptr) == NULL){
		permanent_objs.add(obj_ptr);
		return 0;
	} else {
		return 1;
	}
}

int WorldManager::unsetPermanent(GameObject *obj_ptr){
	permanent_objs.remove(obj_ptr);
	return 0;
}

void WorldManager::update(){
	list<GameObject*>::iterator iter;
	// If ready to clear the list, clear the list (except for the saved objects)
	if(clear){
		iter = game_objs.begin();
		while(iter != game_objs.end()){
			// If the pointer is not in the permanent lists
			if(permanent_objs.find(*iter) == NULL){
				// Mark the object for deletion
				markForDelete(*iter);
			}
			iter++;
		}
		clear = false;
	}

	// Remove all objects marked for deletion
	iter = rm_objs.begin();
	while(iter != rm_objs.end()){
		if((*iter)->isSolid()){
			setPassable((*iter)->getPos().get_x(), (*iter)->getPos().get_y(), true);
		}
		// Remove the pointer from the object list
		removeObj( *iter );
		// Delete the pointer
		delete *iter;
		iter++;
	}
	rm_objs.clear();

	// Send step event to all objects
	EventManager &event_manager = EventManager::getInstance();
	event_manager.sendEvent( new EventStep() );

	// Cause all objects to draw themselves
	// Draw low altitudes first, high altitudes last
	for(int i = (-1 * MAX_ALTITUDE); i <= MAX_ALTITUDE; i++){
		// Draw all objects with an altitude of i
		iter = game_objs.begin();
		while(iter != game_objs.end()){
			if((*iter)->getAltitude() == i){
				(*iter)->draw();
			}
			iter++;
		}
	}
}

int WorldManager::moveObj(GameObject* obj_ptr, Position new_pos){
	// if not a null pointer
	if(obj_ptr){
		Position oldPos = obj_ptr->getPos();
		obj_ptr->setOldPos(oldPos);
		obj_ptr->setPos(new_pos);
		GameObject* collided_obj = findCollision(obj_ptr);
		if(collided_obj){
			obj_ptr->setPos(oldPos);
			// if a collision is detected, move the object back
			// where it was
			// Send the event to both objects
			EventCollision *e = new EventCollision();
			e->setObj1(obj_ptr);
			e->setObj2(collided_obj);
			e->setPos(new_pos);

			obj_ptr->eventHandler(e);

			// By collided_obj's perspective, it is object 1
			e->setObj1(collided_obj);
			e->setObj2(obj_ptr);

			collided_obj->eventHandler(e);
			return 1;
		}
		// determine if the object is on the grid
		int x = obj_ptr->getPos().get_x();
		int y = obj_ptr->getPos().get_y();

		GraphicsManager &graphics_manager = GraphicsManager::getInstance();
		int window_x = graphics_manager.getHorizontal();
		int window_y = graphics_manager.getVertical();

		if(x < 0 || x > world_boundary.getWidth() ||
		   y < 0 || y > world_boundary.getHeight()){
			// if the object is not on the grid, send the event
			obj_ptr->eventHandler(new EventOut());
		}

		if (obj_ptr->isSolid()) {
			setPassable(oldPos.get_x(),oldPos.get_y(), true);
			setPassable(x,y,false);
		}

		return 0;
	} else {
		return 1;
	}
}

bool WorldManager::boxIntersect(Box box1, Box box2){
	// Get the corners for both boxes
	int Ax1 = box1.getPos().get_x();
	int Ax2 = box1.getPos().get_x() + box1.getWidth() - 1;
	int Ay1 = box1.getPos().get_y();
	int Ay2 = box1.getPos().get_y() + box1.getHeight() - 1;

	int Bx1 = box2.getPos().get_x();
	int Bx2 = box2.getPos().get_x() + box2.getWidth() - 1;
	int By1 = box2.getPos().get_y();
	int By2 = box2.getPos().get_y() + box2.getHeight() - 1;

	// Determine if there is overlap
	if( ((Bx1 <= Ax1 && Ax1 <= Bx2) ||
	    (Ax1 <= Bx1 && Bx1 <= Ax2)) &&
	    ((By1 <= Ay1 && Ay1 <= By2) ||
	    (Ay1 <= By1 && By1 <= Ay2))){
		return true;
	} else {
		return false;
	}
}

int WorldManager::testCollision(Position obj_pos){
	Box obj_box(obj_pos, 1, 1);
	// Iterate through all the objects to find the first collision, if any
	list<GameObject*>::iterator iter = game_objs.begin();
	while(iter != game_objs.end()){
		GameObject* temp_obj = *iter;

		// If a box intersects with this position, there is a collision
		if(boxIntersect(obj_box, temp_obj->getBox())){
			return true;
		}
		iter++;
	}
	return false;
}


GameObject* WorldManager::findCollision(GameObject* obj_ptr){
	// If the object is not solid, it won't collide with anything
	if(!obj_ptr->isSolid()){
		return NULL;
	}

	// The position to compare all other positions to
	Position temp_pos = obj_ptr->getPos();
	// Iterate through all the objects to find the first collision, if any
	list<GameObject*>::iterator iter = game_objs.begin();
	while(iter != game_objs.end()){
		// If the positions are equal, the object is not colliding with itself,
		// and the other object is also solid, return the collision
		GameObject* temp_obj = *iter;

		if(temp_obj->isSolid() &&
		   boxIntersect(obj_ptr->getBox(), temp_obj->getBox()) &&
		   temp_obj != obj_ptr){
			return temp_obj;
		}
		iter++;
	}
	return NULL;
}

bool WorldManager::isPassable(int x, int y) {
//	LogManager &log_manager = LogManager::getInstance();
//	log_manager.writeLog(1,"WorldManager::isPassable(): checking %d %d", x,y);
	if ((x >= 0 && x < MAP_WIDTH)&&(y >= 0 && y < MAP_HEIGHT)) {
//		log_manager.writeLog(1,"WorldManager::isPassable(): within bounds of map.");
		return path_map[x][y];
	}
	return false;
}

void WorldManager::setPassable(int x, int y, bool pass) {
	if ((x >= 0 && x < MAP_WIDTH)&&(y >= 0 && y < MAP_HEIGHT)) {
		path_map[x][y] = pass;
	}
}

void WorldManager::resetPathMap() {
	for (int i = 0; i < MAP_WIDTH; i++) {
		for (int j = 0; j < MAP_HEIGHT; j++) {
			setPassable(i,j,true);
		}
	}
}
