// GameController.cpp
// Brian Seney, Alfredo Porras
// IMGD3000

#include "GameController.h"

#include "Engine/LogManager.h"
#include "Engine/WorldManager.h"
#include "Engine/EventStep.h"
#include "Engine/GraphicsManager.h"
#include "Engine/ResourceManager.h"
#include "Engine/GameManager.h"

#include "Player.h"
#include "Picture.h"
#include "Ball.h"
#include "Opponent.h"
#include "OpponentDrunk.h"
#include "OpponentZombie.h"
#include "OpponentGunZombie.h"
#include "OpponentBig.h"
#include "OpponentSpawner.h"
#include "OpponentBallSteal.h"
#include "OpponentHunter.h"
#include "Boss.h"
#include "Goal.h"
#include "Spawner.h"
#include "EventScore.h"
#include "PreLevelPicture.h"
#include "PostLevelPicture.h"
#include "EndGamePicture.h"
#include "Block.h"
#include "BlockBreakable.h"
#include "BlockMovable.h"
#include "Weapon.h"
#include "EnemyArrow.h"
#include "EventPlayerDeath.h"

#include <fstream>
#include <sstream>
#include <string>
#include <cstring>
#include <math.h>

using namespace std;

void createBall(int x_pos, int y_pos){ // possibly find a way to encapsulate these within the class
	Ball *ball_ptr = new Ball();
	ball_ptr->moveAbsolute(x_pos, y_pos);
}

void createOpponent(int x_pos, int y_pos){
	WorldManager &world_manager = WorldManager::getInstance();
	world_manager.moveObj(new Opponent(), Position(x_pos, y_pos));
}

void createDrunkOpponent(int x_pos, int y_pos){
	WorldManager &world_manager = WorldManager::getInstance();
	world_manager.moveObj(new OpponentDrunk(), Position(x_pos, y_pos));
}

void createZombieOpponent(int x_pos, int y_pos){
	WorldManager &world_manager = WorldManager::getInstance();
	world_manager.moveObj(new OpponentZombie(), Position(x_pos, y_pos));
}

void createGunZombieOpponent(int x_pos, int y_pos){
	WorldManager &world_manager = WorldManager::getInstance();
	world_manager.moveObj(new OpponentGunZombie(), Position(x_pos, y_pos));
}

void createBallStealOpponent(int x_pos, int y_pos){
	WorldManager &world_manager = WorldManager::getInstance();
	world_manager.moveObj(new OpponentBallSteal(), Position(x_pos, y_pos));
}

void createBigOpponent(int x_pos, int y_pos){
	OpponentBig* temp = new OpponentBig();
	// set the initial position
	temp->setPos(x_pos, y_pos);
	WorldManager &world_manager = WorldManager::getInstance();
	// if object could not be moved, delete it
	if(world_manager.moveObj(temp, Position(x_pos - 1, y_pos - 1)))
		world_manager.markForDelete(temp);
}

void createSpawnerOpponent(int x_pos, int y_pos){
	OpponentSpawner* temp = new OpponentSpawner();
	// set the initial position
	temp->setPos(x_pos, y_pos);
	WorldManager &world_manager = WorldManager::getInstance();
	// if object could not be moved, delete it
	if(world_manager.moveObj(temp, Position(x_pos - 2, y_pos - 2)))
		world_manager.markForDelete(temp);
}

void createHunterOpponent(int x_pos, int y_pos) {
	WorldManager &world_manager = WorldManager::getInstance();
	world_manager.moveObj(new OpponentHunter(), Position(x_pos, y_pos));
}

void createArrows(int x_pos, int y_pos){
	new EnemyArrow(x_pos - 1, y_pos, -1, 0);
	new EnemyArrow(x_pos + 1, y_pos, 1, 0);
	new EnemyArrow(x_pos, y_pos - 1, 0, -1);
	new EnemyArrow(x_pos, y_pos + 1, 0, 1);
}

GameController::GameController(){
	altitude = 2; // keep the display above everything else

	current_level = 0;

	WorldManager &world_manager = WorldManager::getInstance();
	world_manager.addObj(this);
	world_manager.setPermanent(this); // object will be permanent

	EventManager &event_manager = EventManager::getInstance();
	event_manager.registerInterest(this, STEP_EVENT);
	event_manager.registerInterest(this, SCORE_EVENT);
	event_manager.registerInterest(this, PLAYER_DEATH_EVENT);

	// Create two player objects
	player1 = new Player();
	player1->setColor(1);

	player2 = new Player();
	player2->setColor(4);

	// Give the players a starting weapon
	player1->addWeapon( SWORD_TYPE, ' ', player1->getColor() * 8, 0);
	player2->addWeapon( SWORD_TYPE, ' ', player2->getColor() * 8, 0);

	// Make the players permananet objects
	world_manager.setPermanent(player1);
	world_manager.setPermanent(player2);
}

GameController::~GameController(){
	EventManager &event_manager = EventManager::getInstance();
	event_manager.removeInterest(this, STEP_EVENT);
	event_manager.removeInterest(this, SCORE_EVENT);
	event_manager.removeInterest(this, PLAYER_DEATH_EVENT);
}

void GameController::startGame(int game_type){
	game_mode = game_type;
	// Load the level strings into the game
	// Set the key bindings for the players
	if(game_mode == 1){
		level_pictures.push_back("Sprites/PreLevel1.txt");
		player1->setKeys( 0, KEY_UP, KEY_DOWN, KEY_LEFT, KEY_RIGHT, 'z', 'c', 'x');
	} else {
		level_pictures.push_back("Sprites/PreLevel1(2Player).txt");
		player1->setKeys( 0, KEY_UP, KEY_DOWN, KEY_LEFT, KEY_RIGHT, ',', '/', '.');
	}
	player2->setKeys( 0, 'w', 's', 'a', 'd', '1', '3', '2');

	level_pictures.push_back("Sprites/PreLevel2.txt");
	level_pictures.push_back("Sprites/PreLevel3.txt");
	level_pictures.push_back("Sprites/PreLevel4.txt");
	level_pictures.push_back("Sprites/PreLevel5.txt");
	level_pictures.push_back("Sprites/PreLevel6.txt");
	level_pictures.push_back("Sprites/PreLevel7.txt");
	level_pictures.push_back("Sprites/PreLevel8.txt");
	level_pictures.push_back("Sprites/PreLevel9.txt");
	level_pictures.push_back("Sprites/PreLevel10.txt");
	level_pictures.push_back("Sprites/PreBossLevel.txt");

	level_files.push_back("Levels/Level1.txt");
	level_files.push_back("Levels/Level2.txt");
	level_files.push_back("Levels/Level3.txt");
	level_files.push_back("Levels/Level4.txt");
	level_files.push_back("Levels/Level5.txt");
	level_files.push_back("Levels/Level6.txt");
	level_files.push_back("Levels/Level7.txt");
	level_files.push_back("Levels/Level8.txt");
	level_files.push_back("Levels/Level9.txt");
	level_files.push_back("Levels/Level10.txt");
	level_files.push_back("Levels/BossLevel.txt");

	createLevel(level_pictures[current_level], level_files[current_level]);
}

int GameController::createLevel(string sprite_file, string level_name){
	// hide the gamecontroller display
	is_visible = false;
	player1->setEnabled(false);
	player2->setEnabled(false);

	PreLevelPicture *level_ptr = new PreLevelPicture(sprite_file, level_name);
	return 0;
}

int GameController::loadLevel(string level_name){
	// make the display visible
	is_visible = true;

	// Initialize the fields for the newly loaded level
	level_time = 0;

	player1->setHealth(10);
	player2->setHealth(10);

	WorldManager &world_manager = WorldManager::getInstance();
	world_manager.resetPathMap();
	loadLevelObjects(level_name);
	return 0;
}

int GameController::loadLevelObjects(string level_name){
	// Load the level from the file
	vector<char> level_objects;
	int width = 0;

	ifstream level_file(level_name.c_str());
	if(level_file.is_open()){
		// while the file has input, read in each line and add the integers to the vector
		string temp_str;
		while(level_file.good()){
			getline(level_file, temp_str);
			char temp_char[5000];
			strcpy(temp_char, temp_str.c_str());
			// tokenize the string and add the integers to the vector
			char* temp = strtok(temp_char, " \n");
			while(1){
				if(temp){
					level_objects.push_back( *temp );
				} else {
					break;
				}
				temp = strtok(NULL, " \n");
			}
			// if the width has not been initialized yet, set it
			if(width == 0)
				width = level_objects.size();
		}
	}

	// Now that the positions for each object are known, load them into the world manager
	for(int i = 0; i < level_objects.size(); i++){
		int x_pos = i % width;
		int y_pos = (int)floor(i / width);
		switch( level_objects[i] ){
			case '0': // background: no object created
				break;
			case '.': // also background: no object created
				break;
			case '1': // player 1
			{
				WorldManager &world_manager = WorldManager::getInstance();
				world_manager.moveObj(player1, Position(x_pos, y_pos));
				player1->setEnabled(true);

				if(game_mode == 2){
					WorldManager &world_manager = WorldManager::getInstance();
					world_manager.moveObj(player2, Position(x_pos, y_pos + 2));
					player2->setEnabled(true);
				}
				break;
			}
			case '2': // ball
			{
				createBall(x_pos, y_pos);
				break;
			}
			case '3': // ball spawner
			{
				void (*temp_fp)(int, int) = createBall;
				Spawner *spawn_ptr = new Spawner( temp_fp );
				spawn_ptr->setPos(x_pos, y_pos);
				spawn_ptr->setScoreSpawn(true);
				// start with a ball in the ball spawner
				createBall(x_pos, y_pos);
				break;
			}
			case '4': // goal
			{
				Goal *goal_ptr = new Goal();
				goal_ptr->setPos(x_pos, y_pos);
				break;
			}
			case '5': // block
				new Block(Position(x_pos, y_pos));
				break;
			case '6': // breakable block
				new BlockBreakable(Position(x_pos, y_pos), 5);
				break;
			case '7': // movable block
				new BlockMovable(Position(x_pos, y_pos));
				break;
			case '8': // health pickup
			{ // internally represented as a stationary bullet
				Bullet* health_ptr = new Bullet();
				health_ptr->setPos(x_pos, y_pos);
				health_ptr->setChar('+');
				health_ptr->setColor(57);
				health_ptr->setType( HEALTH_TYPE );

				WorldManager &world_manager = WorldManager::getInstance();
				world_manager.addObj(health_ptr);
				break;
			}
			case '9': // arrow spawner
			{
				void (*temp_fp)(int, int) = createArrows;
				Spawner *spawn_ptr = new Spawner( temp_fp );
				spawn_ptr->setPos(x_pos, y_pos);
				spawn_ptr->setSpawnSlowdown(60);
				spawn_ptr->setScoreSpawn(false);
				break;
			}
			case 'a': // basic opponent
				createOpponent(x_pos, y_pos);
				break;
			case 'b': // drunk opponent
				createDrunkOpponent(x_pos, y_pos);
				break;
			case 'c': // zombie
				createZombieOpponent(x_pos, y_pos);
				break;
			case 'd': // gun zombie
				createGunZombieOpponent(x_pos, y_pos);
				break;
			case 'e': // big zombie
				createBigOpponent(x_pos, y_pos);
				break;
			case 'f': // spawner zombie
				createSpawnerOpponent(x_pos, y_pos);
				break;
			case 'g': // ball steal opponent
				createBallStealOpponent(x_pos, y_pos);
				break;
			case 'h':
				createHunterOpponent(x_pos,y_pos);
				break;
			case 'A': // basic opponent spawner
			{
				void (*temp_fp)(int, int) = createOpponent;
				Spawner *spawn_ptr = new Spawner( temp_fp );
				spawn_ptr->setPos(x_pos, y_pos);
				spawn_ptr->setSpawnSlowdown(200);
				spawn_ptr->setScoreSpawn(false);
				break;
			}
			case 'B': // drunk opponent spawner
			{
				void (*temp_fp)(int, int) = createDrunkOpponent;
				Spawner *spawn_ptr = new Spawner( temp_fp );
				spawn_ptr->setPos(x_pos, y_pos);
				spawn_ptr->setSpawnSlowdown(200);
				spawn_ptr->setScoreSpawn(false);
				break;
			}
			case 'C': // zombie opponent spawner
			{
				void (*temp_fp)(int, int) = createZombieOpponent;
				Spawner *spawn_ptr = new Spawner( temp_fp );
				spawn_ptr->setPos(x_pos, y_pos);
				spawn_ptr->setSpawnSlowdown(200);
				spawn_ptr->setScoreSpawn(false);
				break;
			}
			case 'D': // gun zombie opponent spawner
			{
				void (*temp_fp)(int, int) = createGunZombieOpponent;
				Spawner *spawn_ptr = new Spawner( temp_fp );
				spawn_ptr->setPos(x_pos, y_pos);
				spawn_ptr->setSpawnSlowdown(200);
				spawn_ptr->setScoreSpawn(false);
				break;
			}
			case 'E': // big zombie opponent spawner
			{
				void (*temp_fp)(int, int) = createBigOpponent;
				Spawner *spawn_ptr = new Spawner( temp_fp );
				spawn_ptr->setPos(x_pos, y_pos);
				spawn_ptr->setSpawnSlowdown(300);
				spawn_ptr->setScoreSpawn(false);
				break;
			}
			case 'F': // spawner opponent spawner
			{
				void (*temp_fp)(int, int) = createSpawnerOpponent;
				Spawner *spawn_ptr = new Spawner( temp_fp );
				spawn_ptr->setPos(x_pos, y_pos);
				spawn_ptr->setSpawnSlowdown(400);
				spawn_ptr->setScoreSpawn(false);
				break;
			}
			case 'G': // ball steal opponent spawner
			{
				void (*temp_fp)(int, int) = createBallStealOpponent;
				Spawner *spawn_ptr = new Spawner( temp_fp );
				spawn_ptr->setPos(x_pos, y_pos);
				spawn_ptr->setSpawnSlowdown(400);
				spawn_ptr->setScoreSpawn(false);
				break;
			}
			case 'H': // hunter opponent spawner
			{
				void (*temp_fp)(int, int) = createHunterOpponent;
				Spawner *spawn_ptr = new Spawner( temp_fp );
				spawn_ptr->setPos(x_pos, y_pos);
				spawn_ptr->setSpawnSlowdown(400);
				spawn_ptr->setScoreSpawn(false);
				break;
			}
			case 'X': // boss
			{
				Boss *boss_ptr = new Boss();
				boss_ptr->setPos(x_pos, y_pos);
				break;
			}
		}
	}

	return 0;
}

int GameController::eventHandler(Event *event_ptr){
	if(event_ptr->getType().compare(STEP_EVENT) == 0){
		level_time += FRAME_TICKS;
		return 1;
	} else if(event_ptr->getType().compare(SCORE_EVENT) == 0){
		// if scored, check the win condition
		EventScore* score_ptr = static_cast<EventScore*>(event_ptr);
		int score_type = score_ptr->getScoreType();
		if(score_type == 2){
			winLevel();
		}
		return 1;
	} else if(event_ptr->getType().compare(PLAYER_DEATH_EVENT) == 0){
		// Determine if both players, or just one, is dead
		if(!player1->isEnabled() && !player2->isEnabled()){
			loseLevel();
		}
		return 1;
	}
	return 0;
}

void GameController::winLevel(){
	// Hide the display
	is_visible = false;
	player1->setEnabled(false);
	player2->setEnabled(false);
	player1->setHoldingBall(NULL);
	player2->setHoldingBall(NULL);

	WorldManager &world_manager = WorldManager::getInstance();
	world_manager.setClear();

	// Load the next level
	current_level++;
	if(current_level < level_files.size()){
		PostLevelPicture *pic_ptr = new PostLevelPicture( "Sprites/WinLevel.txt", level_pictures[current_level], level_files[current_level]);
		pic_ptr->setPos(10, 5);
		world_manager.setPermanent(pic_ptr);
	} else {
		// If there are no more levels, you win!
		ResourceManager &resource_manager = ResourceManager::getInstance();
		Sprite *sprite_ptr = resource_manager.loadSprite("Sprites/WinGame.txt");
		EndGamePicture *temp = new EndGamePicture(sprite_ptr);
		temp->setPos(18, 3);

		Picture* pic_ptr = new Picture( resource_manager.loadSprite("Sprites/GoalString.txt") );
		pic_ptr->setPos(19, 15);
		pic_ptr->setTimetoLive(-1);

		world_manager.setPermanent(temp);
		world_manager.setPermanent(pic_ptr);
	}
}

void GameController::loseLevel(){
	// Hide the display
	is_visible = false;
	player1->setEnabled(false);
	player2->setEnabled(false);
	player1->setHoldingBall(NULL);
	player2->setHoldingBall(NULL);

	WorldManager &world_manager = WorldManager::getInstance();
	world_manager.setClear();

	// Retry the level
	PostLevelPicture *pic_ptr = new PostLevelPicture( "Sprites/LoseLevel.txt", level_pictures[current_level], level_files[current_level]);
	pic_ptr->setPos(15, 15);
	world_manager.setPermanent(pic_ptr);
}

void GameController::draw(){
	// if not visible, return
	if(!is_visible) return;

	int p1_weapon_pos; // where to draw the current weapons
	int p2_weapon_pos;

	stringstream hud_string (stringstream::in | stringstream::out);

	// The player 1 score display (if player 1 is enabled)
	if(player1->isEnabled()){
		if(player1->getHealth() >= 10)
			hud_string << "Player1:" << player1->getHealth() << "hp";
		else
			hud_string << "Player1: " << player1->getHealth() << "hp";
		hud_string << " Weapon: ";
		p1_weapon_pos = hud_string.str().length() - 1;
	} else {
		hud_string << "                     ";
	}

	// The timer display (convert the nanosecond time_remaining to minute:second)
	hud_string << " ||Time:";
	long int temp_level_time = level_time;
	int minutes = (int)floor(temp_level_time / 60000000);
	temp_level_time -= minutes * 60000000;
	int seconds = (int)floor(temp_level_time / 1000000);
	if(seconds < 10)
		hud_string << minutes << ":0" << seconds << "||";
	else
		hud_string << minutes << ":" << seconds << "||";

	// The second player score display (if the second player is enabled)
	if(player2->isEnabled()){
		if(player2->getHealth() >= 10)
			hud_string << " Player2:" << player2->getHealth() << "hp";
		else
			hud_string << " Player2: " << player2->getHealth() << "hp";
		hud_string << " Weapon: ";
		p2_weapon_pos = hud_string.str().length() - 1;
	}

	char temp[200];
	strcpy(temp, hud_string.str().c_str() );
	GraphicsManager &graphics_manager = GraphicsManager::getInstance();
	graphics_manager.drawStr(Position(0, 0), temp );

	// If player 1 is enabled, draw the player 1 current weapon
	if(player1->isEnabled()){
		Weapon* current_weapon = player1->getCurrentWeapon();
		graphics_manager.drawCh(Position(p1_weapon_pos, 0), current_weapon->getChar(), current_weapon->getColor());
	}
	// If player 2 is enabled, draw the player 2 current weapon
	if(player2->isEnabled()){
		Weapon* current_weapon = player2->getCurrentWeapon();
		graphics_manager.drawCh(Position(p2_weapon_pos, 0), current_weapon->getChar(), current_weapon->getColor());
	}
}
