// Player.cpp
// Brian Seney, Alfredo Porras
// IMGD3000

#include "Engine/EventManager.h"
#include "Engine/EventKeyboard.h"
#include "Engine/EventMouse.h"
#include "Engine/EventCollision.h"
#include "EventBallSteal.h"

#include "Engine/GameManager.h"
#include "Engine/LogManager.h"
#include "Engine/GraphicsManager.h"
#include "Engine/ResourceManager.h"
#include "Engine/WorldManager.h"
#include <string>
#include <curses.h>
#include <stdarg.h>

#include "Player.h"
#include "Ball.h"
#include "Sword.h"
#include "Arrow.h"
#include "Weapon.h"
#include "Opponent.h"
#include "Bomb.h"
#include "Block.h"
#include "BlockMovable.h"
#include "EnemyArrow.h"
#include "EventTargetMovement.h"
#include "EventPlayerDeath.h"

#define BALL_VELOCITY 1.0
#define BALL_ACCEL 0.05

#define PLAYER_BALL_COLOR 6

Player::Player(){
	is_solid = true;
	object_type = PLAYER_TYPE;

	direction = '^';
	holding_ball = NULL;
	color = 1;

	health = 10;
	setEnabled(true);

	current_weapon = 0;
	resolve_move = false;

	addWeapon( ARROW_TYPE, '+', 0, 0); //
	addWeapon( TRIPLE_SHOT_TYPE, '3', 56, 10); //
	addWeapon( BOMB_TYPE, 'Y', 0, 20); //
	addWeapon( TRIPLE_BOMB_TYPE, 'Y', 56, 80); //

	WorldManager &world_manager = WorldManager::getInstance();
	world_manager.addObj(this);

	EventManager &event_manager = EventManager::getInstance();
	event_manager.registerInterest(this, KEYBOARD_EVENT);
	event_manager.registerInterest(this, MOUSE_EVENT);
	event_manager.registerInterest(this, COLLISION_EVENT);
	event_manager.registerInterest(this, BALL_STEAL_EVENT);
}

Player::~Player(){
	EventManager &event_manager = EventManager::getInstance();
	event_manager.removeInterest(this, KEYBOARD_EVENT);
	event_manager.removeInterest(this, MOUSE_EVENT);
	event_manager.removeInterest(this, COLLISION_EVENT);
	event_manager.removeInterest(this, BALL_STEAL_EVENT);
}

int Player::eventHandler(Event *event_ptr){
	if(!is_enabled){ return 0; }

	if(event_ptr->getType().compare(KEYBOARD_EVENT) == 0){
		int key = static_cast<EventKeyboard*>(event_ptr)->getKey();
		// Movement
		if(key == keys[0]){
			direction = '^';
			move(0, -1);
		} else if(key == keys[1]){
			direction = 'v';
			move(0, 1);
		} else if(key == keys[2]){
			direction =  '<';
			move(-1, 0);
		} else if(key == keys[3]){
			direction = '>';
			move(1, 0);
		// Shoot
		} else if(key == keys[4]){
			if(holding_ball){
				if(direction == '^'){
					shootBall(0, -BALL_VELOCITY, 0, BALL_ACCEL);
				} else if(direction == 'v'){
					shootBall(0, BALL_VELOCITY, 0, -BALL_ACCEL);
				} else if(direction == '<'){
					shootBall(-BALL_VELOCITY, 0, BALL_ACCEL, 0);
				} else if(direction == '>'){
					shootBall(BALL_VELOCITY, 0, -BALL_ACCEL, 0);
				}
			} else {
				if(direction == '^'){
					weapons[current_weapon]->
					shootWeapon( pos.get_x(), pos.get_y() - 1, 0, -1, 0, 0 );
				} else if(direction == 'v'){
					weapons[current_weapon]->
					shootWeapon( pos.get_x(), pos.get_y() + 1, 0, 1, 0, 0);
				} else if(direction == '<'){
					weapons[current_weapon]->
					shootWeapon( pos.get_x() - 1, pos.get_y(), -1, 0, 0, 0 );
					} else if(direction == '>'){
					weapons[current_weapon]->
					shootWeapon( pos.get_x() + 1, pos.get_y(), 1, 0, 0, 0 );
					}
				}
		// Drop Ball
		} else if(key == keys[5]){
			holding_ball = NULL;
		// Switch Weapon
		} else if(key == keys[6]){
			current_weapon++;
			if(current_weapon >= weapons.size())
				current_weapon = 0;
		// Escape key: End the game
		} else if(key == 27){
			GameManager &game_manager = GameManager::getInstance();
				game_manager.stop();
		}
		return 1;
	} else if(event_ptr->getType().compare(MOUSE_EVENT) == 0){
		EventMouse *mouse_ptr = static_cast<EventMouse*>(event_ptr);
		int x = mouse_ptr->getMousePosition().get_x();
		int y = mouse_ptr->getMousePosition().get_y();
		return 1;
	} else if(event_ptr->getType().compare(COLLISION_EVENT) == 0){
		EventCollision *collision_ptr = static_cast<EventCollision*>(event_ptr);
		// If colliding with the ball, update the player to have the ball
		GameObject* obj_ptr = collision_ptr->getObj2();
		if(obj_ptr->getType().compare( BALL_TYPE ) == 0 ){
			holding_ball = static_cast<Ball*>(collision_ptr->getObj2());
			// Let other players know to drop the ball
			EventManager &event_manager = EventManager::getInstance();
			event_manager.sendEvent( new EventBallSteal(holding_ball, this) );
		// If colliding with the enemy, destroy the enemy, lower health
		} else if(obj_ptr->getType().compare( OPPONENT_TYPE ) == 0){
			adjustHealth(-1);
		// Health pickup collisions: restore 1 health
		} else if(obj_ptr->getType().compare( HEALTH_TYPE ) == 0){
			adjustHealth(1);
			WorldManager &world_manager = WorldManager::getInstance();
			world_manager.markForDelete(collision_ptr->getObj2());
		// Enemy bullet collisons: take 1 damage
		} else if(obj_ptr->getType().compare( ENEMY_ARROW_TYPE ) == 0){
			adjustHealth(-1);
		// Block Collisions
		} else if (obj_ptr->getType().compare("block") == 0) {
			Block *p_block = static_cast<Block *>(obj_ptr);
			if (p_block->getBlockType().compare("movable") == 0){
				int x = 0;
				int y = 0;
				if(direction == '^')
					y = -1;
				else if(direction == 'v')
					y = 1;
				else if(direction == '<')
					x = -1;
				else if(direction == '>')
					x = 1;
				WorldManager &world_manager = WorldManager::getInstance();
				Position temp_pos = obj_ptr->getPos();
				// If there is not a collision at the new location
				if(!world_manager.testCollision( Position(obj_ptr->getPos().get_x() + x, obj_ptr->getPos().get_y() + y)) ){
					world_manager.moveObj(obj_ptr, Position(temp_pos.get_x() + x, temp_pos.get_y() + y));
					LogManager &log_manager = LogManager::getInstance();
					log_manager.writeLog(1, "X, Y: %i %i", temp_pos.get_x() + x, temp_pos.get_y() + y);
					world_manager.moveObj(this, temp_pos);
				}
			}
		}
		return 1;
	} else if(event_ptr->getType().compare( BALL_STEAL_EVENT ) == 0){
		// If the ball being stolen is this one, release the ball
		EventBallSteal *steal_ptr = static_cast<EventBallSteal*>(event_ptr);
		// If holding ball is not null and is the  ball being stolen
		if(holding_ball && holding_ball == steal_ptr->getStolenBall()){
			// If the player is not stealing the ball from itself
			if(steal_ptr->getStolenPlayer() == this){
			} else {
				holding_ball = NULL;
			}
		}
		return 1;
	}

	return 0;
}

void Player::draw(){
	if(!is_enabled){return;}
	// update the color
	int temp_color;
	if(holding_ball)
		temp_color = (8 * PLAYER_BALL_COLOR) + color;
	else
		temp_color = 8 * color;
	GraphicsManager &graph_manager = GraphicsManager::getInstance();
	graph_manager.drawCh(pos, direction, temp_color);
}

void Player::move(int x, int y){
	WorldManager &world_manager = WorldManager::getInstance();
	if(!holding_ball){
		world_manager.moveObj(this, Position(pos.get_x() + x, pos.get_y() + y));
		if (resolve_move) {
			world_manager.moveObj(this, move_to);
			resolve_move = false;
		}
	} else {
		// if the ball can be moved in front of the position (move returns 0), move
		if(holding_ball->moveAbsolute(pos.get_x() + x, pos.get_y() + y) == 0){
			if(holding_ball->moveAbsolute(pos.get_x() + 2 * x, pos.get_y() + 2 * y) == 0){
				world_manager.moveObj(this, Position(pos.get_x() + x, pos.get_y() + y));
			}
		}
	}
	EventManager &event_manager = EventManager::getInstance();
	event_manager.sendEvent(new EventTargetMovement(pos));
}

void Player::setCurrentWeapon(Weapon* new_weapon){
	for(int i = 0; i < weapons.size(); i++){
		if(weapons[i] == new_weapon){
			current_weapon = i;
			return;
		}
	}
}

void Player::shootBall(float vel_x, float vel_y, float accel_x, float accel_y){\
	if(holding_ball){
		holding_ball->setVelocity(vel_x, vel_y);
		holding_ball->setAccel(accel_x, accel_y);

		holding_ball = NULL;
	}
}

void Player::adjustHealth(int new_health){
	health += new_health;

	// Display the damage sprite
	ResourceManager &resource_manager = ResourceManager::getInstance();
	Picture* pic_ptr = new Picture();
	pic_ptr->setSprite( resource_manager.loadSprite("Sprites/Damage.txt"));
	pic_ptr->setPos(pos.get_x() - 1, pos.get_y() - 1);
	pic_ptr->setAltitude(2);
	pic_ptr->setTimetoLive(4);
	
	if(health <= 0){
		// send death event
		setEnabled(false);
		EventManager &event_manager = EventManager::getInstance();
		event_manager.sendEvent( new EventPlayerDeath() );
	}
}

void Player::addWeapon(string weapon_type, char display_char, int weapon_color, int slowdown){
	// if the player already has the weapon, return
	for(int i = 0; i < weapons.size(); i++){
		if(weapons[i]->getWeaponType().compare( weapon_type ) == 0){
			return;
		}
	}
	Weapon* weapon_ptr = new Weapon();
	weapon_ptr->setWeaponType( weapon_type );
	weapon_ptr->setChar( display_char );
	weapon_ptr->setColor( weapon_color );
	weapon_ptr->setSlowdown( slowdown );
	weapons.push_back(weapon_ptr);
}

void Player::setKeys( int first, ...){
	va_list player_keys;
	va_start(player_keys, first);
	for(int i = 0; i < 7; i++){
		keys.push_back( va_arg(player_keys, int) );
	}
}
